from fiql_parser import Expression, Constraint, Operator, parse_str_to_expression

from dateutil.relativedelta import relativedelta
import datetime

def transform_yearid(yearid, selector):
    if selector == 'yearid':
        return yearid
    elif selector == 'termid':
        return yearid * 100
    elif 'date' in selector:
        academic_year = yearid + 1990
        return datetime.date(academic_year, 7, 1)

def get_constraint_rules(selector, yearid=None):
    if selector == 'yearid':
        return {'step_size': 1, 'stop': 10}
    elif selector == 'termid':
        termid = transform_yearid(yearid, 'termid')
        return {'step_size': 100, 'stop': -termid}
    elif 'date' in selector:
        return {'step_size': relativedelta(years=1), 'stop': datetime.date(2000, 7, 1)}
    else:
        return {'step_size': 10000, 'stop': 1}

def get_constraint_values(selector, arg_value, step_size):
    if selector == 'yearid':
        arg_next = arg_value + step_size
    elif selector == 'termid' and arg_value < 0:
        arg_next = arg_value - step_size
    elif 'date' in selector and type(arg_value) is str:
        arg_value = datetime.datetime.strptime(arg_value, '%Y-%m-%d').date()
        arg_next = arg_value + step_size
    else:
        arg_next = None
    return {'start': arg_value, 'end': arg_next}

def get_query_expression(selector, start, end):
    query_expression = Expression()
    if type(start) is int and start < 0:
        query_expression.add_element(Constraint(selector, '=gt=', str(end)))
        query_expression.add_element(Operator(';'))
        query_expression.add_element(Constraint(selector, '=le=', str(start)))
    elif end is None:
        query_expression.add_element(Constraint(selector, '=ge=', str(start)))
    else:
        query_expression.add_element(Constraint(selector, '=ge=', str(start)))
        query_expression.add_element(Operator(';'))
        query_expression.add_element(Constraint(selector, '=lt=', str(end)))
    return str(query_expression)

def generate_historical_queries(current_yearid, query_constraint_selector):
    ## transform yearid to constraint value
    max_constraint_value = transform_yearid(current_yearid, query_constraint_selector)

    ## get step and stoppage critera for constraint type
    constraint_rules = get_constraint_rules(query_constraint_selector, current_yearid)
    stop_constraint_value = constraint_rules['stop']
    constraint_step_size = constraint_rules['step_size']

    ## generate probing queries
    working_constraint_value = max_constraint_value
    probing_query_expressions = []
    while working_constraint_value >= stop_constraint_value:
        constraint_values = get_constraint_values(query_constraint_selector, working_constraint_value, constraint_step_size)
        query_expression = get_query_expression(query_constraint_selector, **constraint_values)
        probing_query_expressions.append(query_expression)
        working_constraint_value = working_constraint_value - constraint_rules['step_size']
    return probing_query_expressions

def parse_fiql_selector(query_string):
    # parse query string to get selector
    query_expression = parse_str_to_expression(query_string)
    query_constraint = query_expression.elements[0]
    return query_constraint.selector



__version__ = '1.0.2'

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
from requests.auth import HTTPBasicAuth
import requests

from datetime import datetime
from collections import namedtuple
import math

class PowerSchool:
    """
    """
    def __init__(self, host):
        """
        """
        self.host = host
        self.base_url = f'https://{self.host}'
        self.access_token = None
        self.metadata = None
        self.session = requests.Session()
        self.session.headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        }

    def _request(self, method, path, params={}, data={}):
        """
        """
        url = f'{self.base_url}{path}'
        try:
            response = self.session.request(method, url=url, params=params, json=data)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as e:
            print(e)
            try:
                response_dict = response.json()
                print(f"\t{response_dict.get('message')}")
                for er in response_dict.get('errors'):
                    print(f"\t\t{er.get('resource')}: {er.get('field')} - {er.get('code')}")
            except:
                pass
            raise e

    def _metadata(self):
        """
        """
        path = '/ws/v1/metadata'
        response_dict = self._request('GET', path)

        metadata_dict = response_dict.get('metadata')
        MetadataTuple = namedtuple('Metadata', sorted(metadata_dict))
        return MetadataTuple(**metadata_dict)

    def authorize(self, **kwargs):
        """
        """
        access_token = kwargs.get('access_token')
        client_credentials = kwargs.get('client_credentials')

        # check if access token supplied
        if access_token:
            # check if access token is still valid
            expires_at = datetime.fromtimestamp(access_token.get('expires_at'))
            now = datetime.now()
            if expires_at > now:
                self.access_token = access_token
                self.session.headers['Authorization'] = f"Bearer {self.access_token.get('access_token')}"
                self.metadata = self._metadata()
                return "Authorized!"
            else:
                return "Access token expired!"

        # check for client credentials (tuple)
        if isinstance(client_credentials, tuple):
            client_id, client_secret = client_credentials

            # fetch new access token
            print("Fetching new access token...")
            token_url = f'{self.base_url}/oauth/access_token/'
            auth = HTTPBasicAuth(client_id, client_secret)
            client = BackendApplicationClient(client_id=client_id)
            session = OAuth2Session(client=client)

            token_dict = session.fetch_token(token_url=token_url, auth=auth)

            self.access_token = token_dict
            self.session.headers['Authorization'] = f"Bearer {self.access_token.get('access_token')}"
            self.metadata = self._metadata()
            return "Authorized!"
        else:
            # exit - prompt for credientials tuple
            raise Exception("You must provide a valid access token file or client credentials.")

    def get_schema_table(self, table_name):
        """
        """
        self.schema_table = Schema(self, 'table', table_name)
        return self.schema_table

    def get_named_query(self, query_name):
        """
        """
        self.named_query = Schema(self, 'query', query_name)
        return self.named_query

class Schema:
    """
    """
    def __init__(self, client, schema_type, name):
        self.client = client
        self.schema_type = schema_type
        self.name = name

        if self.schema_type == 'query':
            self.method = 'POST'
        else:
            self.method = 'GET'

    def count(self, body={}, **params):
        """
        """
        path = f'/ws/schema/{self.schema_type}/{self.name}/count'
        filtered_params = { k: params.get(k) for k in ['q', 'students_to_include', 'teachers_to_include'] }
        count_response_dict = self.client._request(self.method, path, filtered_params, body)
        return count_response_dict.get('count')

    def metadata(self, **params):
        path = f'/ws/schema/{self.schema_type}/{self.name}/metadata'
        if self.schema_type == 'query':
            return {}
        else:
            return self.client._request('GET', path, params=params)

    def query(self, dcid=None, body={}, **params):
        """
        """
        path = f'/ws/schema/{self.schema_type}/{self.name}'
        page_size = params.get('pagesize')
        projection = params.get('projection')

        if self.schema_type == 'query':
            pass
        elif projection is None:
            metadata_params = {'expansions': 'access'}
            metadata = self.metadata(**metadata_params)
            columns = metadata.get('columns')
            star_projection = ','.join(
                [ c.get('name').lower() for c in columns if c.get('access') != 'NoAccess' ]
            )
            params.update({'projection': star_projection})
        else:
            params.update({'projection': projection})

        if dcid:
            path = f'{path}/{dcid}'
            filtered_params = { k: params.get(k) for k in ['projection'] }
            response_dict = self.client._request('GET', path, filtered_params)
            return [response_dict.get('tables').get(self.name)]
        else:
            count = self.count(body, **params)
            if count > 0:
                if page_size is None and self.schema_type == 'query':
                    page_size = 0
                elif page_size is None:
                    page_size = self.client.metadata.schema_table_query_max_page_size
                params.update({'pagesize': page_size})

                if page_size is None and self.schema_type == 'query':
                    pages = 1
                elif page_size == 0:
                    pages = 1
                else:
                    pages = math.ceil(count / page_size)

                results = []
                for p in range(pages):
                    params.update({'page': p + 1})
                    response_dict = self.client._request(self.method, path, params, body)
                    for r in response_dict.get('record'):
                        if self.schema_type == 'query':
                            results.append(r)
                        else:
                            results.append(r.get('tables').get(self.name))
                return results



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['powerschool']

package_data = \
{'': ['*']}

install_requires = \
['fiql_parser>=0.15,<0.16',
 'oauthlib>=3.1.0,<4.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'requests>=2.24.0,<3.0.0',
 'requests_oauth>=0.4.1,<0.5.0']

setup_kwargs = {
    'name': 'powerschool',
    'version': '1.0.2',
    'description': 'powerschool is a Python client for the PowerSchool API',
    'long_description': "# PowerSchool\npowerschool is a Python client for the [PowerSchool SIS](https://www.powerschool.com/solutions/student-information-system/powerschool-sis) API\n\n## Installation\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install powerschool.\n```bash\npip install powerschool\n```\n\n## Getting Started\n1. Ensure you have a valid [plugin](https://support.powerschool.com/developer/#/page/plugin-xml) installed with the proper data access provisioned for your purposes.\n   \n2. Instantiate the client by passing the host name of your server.\n    ```python\n    import powerschool\n    ps = powerschool.PowerSchool('my.host.name')\n    ```\n\n3. Authorize the client using:\n    - client credentials (tuple)\n    ```python\n    my_credentials = (client_id, client_secret)\n    ps.authorize(client_credentials=my_credentials)\n    ```\n\n    - a previously saved access token (dict)\n    ```python\n    with open(token_file, 'r') as f:\n        my_token = json.load(f)\n    \n    ps.authorize(access_token=my_token)\n    ```\n\n## Usage\n>*Refer to the [docs](https://support.powerschool.com/developer/#/page/data-access) for full functionality, including resources, searching, and pagination.*\n\n**Instantiate a table or PowerQuery object:**\n```python\nschools_table = ps.get_schema_table('schools')\n\npowerquery = ps.get_named_query('com.pearson.core.student.search.get_student_basic_info')\n```\n\n**Get the record count for a table:**\n```python\nschools_table.count()\n```\n\n**Query all records, all columns on a table:**\n>*Pagination is handled automatically by the client. However, you can manually pass `pagesize` and `page` parameters, should you choose.*\n```python\nschools_table.query()\n```\n\n**Query all records on a table, with filter and columns list:**\n```python\nparams = {\n    'q': 'id=ge=10000',\n    'projection': 'school_number,abbreviation',\n}\nschools_table.query(**params)\n```\n\n**Query a specific record on a table:**\n```python\nschools_table.query(dcid=123)\n```\n\n**Execute a PowerQuery, passing arguments in the body:**\n```python\npayload = {\n    'studentdcid': '5432',\n}\npowerquery.query(body=payload)\n```\n\n## Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## Notice\nPowerSchool® is a registered trademark in the U.S. and/or other countries owned by PowerSchool Education, Inc. or its affiliates. PowerSchool® is used under license.\n",
    'author': 'Charlie Bini',
    'author_email': 'cbini87@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/TEAMSchools/powerschool',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
